/*
 * iostat: report I/O statistics
 * (C) 1998-1999 by Sebastien GODARD (sebastien.godard@wanadoo.fr)
 * 
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                 *
 ***************************************************************************
 */

/* Version */
#define VERSION		"1.1"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>	/* for S_IRUSR, S_IWUSR, etc. */

#ifdef USE_NLS

#include <libintl.h>
#define _(string) gettext(string)

#else

#define _(string) (string)

#endif


#define STAT		"/proc/stat"
#define UPTIME		"/proc/uptime"

#define F_CPU		0x01
#define F_DISK		0x02
#define F_DISK_RBLK	0x04
#define F_DISK_WBLK	0X08
#ifdef TTY
#define F_TTY		0x10
#define F_MASK		0x1f
#else
#define F_MASK		0x0f
#endif


#define NR_DISK		4

#define D_CPU_ONLY	1
#define D_DISK_ONLY	2


unsigned int cpu_user[2], cpu_nice[2], cpu_system[2];
#ifdef CPU_WAIT
unsigned int cpu_wait[2];
#endif
unsigned long int cpu_idle[2];
unsigned int dk_drive[2][NR_DISK];
unsigned int dk_drive_rblk[2][NR_DISK], dk_drive_wblk[2][NR_DISK];
unsigned long int uptime_h, idle_h, uptime[2];
unsigned int uptime_l, idle_l;
#ifdef TTY
unsigned long int tty_in[2], tty_out[2];
#endif
long int interval = 0;


/*
 * Print usage and exit
 */
void usage(char *progname)
{
   fprintf(stderr, _("%s version %s\n"
	           "Usage: %s [-c | -d] [-f <filename>] [interval [count]]\n"),
	   progname, VERSION, progname);
   exit(1);
}


/* 
 * SIGALRM signal handler
 */
void alarm_handler(int sig)
{
   signal(SIGALRM, alarm_handler);
   alarm(interval);
}


/*
 * Initialize stat counters
 */
void init_stat(int curr)
{
   int disk_index;
   
   
   cpu_user[curr] = cpu_nice[curr] = cpu_system[curr] = 0;
#ifdef CPU_WAIT   
   cpu_wait[curr] = 0;
#endif   
   cpu_idle[curr] = 0;
   
#ifdef TTY
   tty_in[curr] = tty_out[curr] = 0;
#endif
   
   for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
      dk_drive[curr][disk_index] = 0;

      dk_drive_rblk[curr][disk_index] = dk_drive_wblk[curr][disk_index] = 0;
   }
   
   uptime[curr] = 0;
}


/*
 * Read machine uptime
 */
void read_uptime(int curr)
{
   FILE *uptimefp;
   char line[80];

   /* Open uptime file */
   if ((uptimefp = fopen(UPTIME, "r")) == NULL) {
      perror("fopen");
      exit(1);
   }
   
   if (fgets(line, 80, uptimefp) == NULL) {
      fprintf(stderr, _("Cannot read %s file\n"), UPTIME);
      exit(1);
   }

   /*
    * Read the uptime of the system, and the amount of time spent in the idle process
    * (expressed in seconds)
    */
   sscanf(line, "%lu.%u %lu.%u", &uptime_h, &uptime_l, &idle_h, &idle_l);

   /* Close stat file */
   fclose(uptimefp);
   
   /* Compute the uptime of the system in jiffies (1/100ths of a second) */
   uptime[curr] = uptime_h * 100 + uptime_l;
}


/*
 * Read stats...
 * (see linux source file linux/fs/proc/array.c)
 */
void read_stat(int curr)
{
   FILE *statfp;
   int sflag = 0;
   char resource_name[16];
   char line[80];
   
   
   /* Open stat file */
   if ((statfp = fopen(STAT, "r")) == NULL) {
      perror("fopen");
      exit(1);
   }

   while (sflag != F_MASK) {

      if (fgets(line, 80, statfp) == NULL) {
	 fprintf(stderr, _("Cannot read %s file\n"), STAT);
	 exit(1);
      }
      
      if (!strncmp(line, "cpu", 3)) {
	 /* Read the number of jiffies spent in user, nice, system and idle mode */
#ifndef CPU_WAIT	 
	 sscanf(line, "%s  %u %u %u %lu",
	        resource_name, &cpu_user[curr], &cpu_nice[curr],
		&cpu_system[curr], &cpu_idle[curr]);
#else	 
	 sscanf(line, "%s  %u %u %u %lu %u",
	        resource_name, &cpu_user[curr], &cpu_nice[curr],
		&cpu_system[curr], &cpu_idle[curr],
		&cpu_wait[curr]);
#endif
	 sflag |= F_CPU;
      }

      else if (!strncmp(line, "disk_rblk", 9)) {
	 /*
	  * Read the number of blocks read from disk (one block is 1024 bytes).
	  * A quick glance at the linux kernel source file linux/drivers/block/ll_rw_blk.c
	  * once made me think this was a number of sectors. Yet, it seems to be
	  * a number of kilobytes... Please tell me if I'm wrong.
	  */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_rblk[curr][0], &dk_drive_rblk[curr][1],
		&dk_drive_rblk[curr][2], &dk_drive_rblk[curr][3]);
	 sflag |= F_DISK_RBLK;
      }
      
      else if (!strncmp(line, "disk_wblk", 9)) {
	 /* Read the number of blocks written to disk */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_wblk[curr][0], &dk_drive_wblk[curr][1],
		&dk_drive_wblk[curr][2], &dk_drive_wblk[curr][3]);
	 sflag |= F_DISK_WBLK;
      }

      /*
       * Note the trailing space here...
       * It is necessary if we don't want to confuse this line with another one
       * (disk_rio or disk_wio for example)
       */
      else if (!strncmp(line, "disk ", 5)) {
	 /* Read tne number of I/O done since the last reboot */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive[curr][0], &dk_drive[curr][1],
		&dk_drive[curr][2], &dk_drive[curr][3]);
	 sflag |= F_DISK;
      }
#ifdef TTY
      else if (!strncmp(line, "tty", 3)) {
	 /* Read the number of characters read and written for all the ttys */
	 sscanf(line, "%s %lu %lu",
		resource_name, &tty_in[curr], &tty_out[curr]);
	 sflag |= F_TTY;
      }
#endif      
   }
   
   /* Close stat file */
   fclose(statfp);
}


/*
 * Print everything now (stats and uptime)
 * The formula used here to display the decimal part is taken from
 * the linux kernel source file linux/fs/proc/array.c
 */
void print_stat(int curr, int cdf)
{
   int disk_index;
   
   
   uptime[!curr] = uptime[curr] - uptime[!curr];

   if (cdf != D_DISK_ONLY) {

#ifdef TTY      
      printf(_("tty:     tin       tout   "));
#endif
      
      printf(_("avg-cpu:  %%user   %%nice    %%sys   %%idle"));
   
#ifdef CPU_WAIT
      printf(_("  %%iowait"));
#endif

      printf("\n");
   
   
#ifdef TTY	     
      printf("   %9lu  %9lu   ",
	     (tty_in[curr] - tty_in[!curr]), (tty_out[curr] - tty_out[!curr]));
#endif	     
      
      printf(_("         %3lu.%02lu  %3lu.%02lu  %3lu.%02lu  %3lu.%02lu   "),
	     ((cpu_user[curr] - cpu_user[!curr]) * 100) / uptime[!curr],
	     (((((cpu_user[curr] - cpu_user[!curr]) * 100) % uptime[!curr]) * 100) / uptime[!curr]) % 100,
	     ((cpu_nice[curr] - cpu_nice[!curr]) * 100) / uptime[!curr],
	     (((((cpu_nice[curr] - cpu_nice[!curr]) * 100) % uptime[!curr]) * 100) / uptime[!curr]) % 100,
	     ((cpu_system[curr] - cpu_system[!curr]) * 100) / uptime[!curr],
	     (((((cpu_system[curr] - cpu_system[!curr]) * 100) % uptime[!curr]) * 100) / uptime[!curr]) % 100,
	     ((cpu_idle[curr] - cpu_idle[!curr]) * 100) / uptime[!curr],
	     (((((cpu_idle[curr] - cpu_idle[!curr]) * 100) % uptime[!curr]) * 100) / uptime[!curr]) % 100);
      
#ifdef CPU_WAIT
      printf(_("%3lu.%02lu"),
	     ((cpu_wait[curr] - cpu_wait[!curr]) * 100) / uptime[!curr],
	     (((((cpu_wait[curr] - cpu_wait[!curr]) * 100) % uptime[!curr]) * 100) / uptime[!curr]) % 100);
#endif
      
      printf("\n");
   }
   
   if (cdf != D_CPU_ONLY) {
   
      printf(_("Disks:     tps    Kb_read/s    Kb_wrtn/s    Kb_read    Kb_wrtn\n"));
   
      for (disk_index = 0; disk_index < NR_DISK; disk_index++) {
         printf(_("hdisk%d  %3lu.%02lu     %5lu.%02lu     %5lu.%02lu    %7u    %7u\n"),
	        disk_index,
	        ((dk_drive[curr][disk_index] - dk_drive[!curr][disk_index]) * 100)
		/ uptime[!curr],
	        (((((dk_drive[curr][disk_index] - dk_drive[!curr][disk_index]) * 100)
		   % uptime[!curr]) * 100) / uptime[!curr]) % 100,
	        ((dk_drive_rblk[curr][disk_index] - dk_drive_rblk[!curr][disk_index]) * 100)
		 / uptime[!curr],
	        (((((dk_drive_rblk[curr][disk_index] - dk_drive_rblk[!curr][disk_index]) * 100)
		    % uptime[!curr]) * 100) / uptime[!curr]) % 100,
	        ((dk_drive_wblk[curr][disk_index] - dk_drive_wblk[!curr][disk_index]) * 100)
		 / uptime[!curr],
	        (((((dk_drive_wblk[curr][disk_index] - dk_drive_wblk[!curr][disk_index]) * 100)
		    % uptime[!curr]) * 100) / uptime[!curr]) % 100,
	        (dk_drive_rblk[curr][disk_index] - dk_drive_rblk[!curr][disk_index]),
	        (dk_drive_wblk[curr][disk_index] - dk_drive_wblk[!curr][disk_index]));
      }
   }
   printf("\n");
}


/*
 * Main entry to the program
 */
int main(int argc, char **argv)
{
   int cdf = 0;
   int i = 0;
   int opt = 1;
   int curr = 1;
   int fd;
   long int count = 1;
   char *filename = NULL;
   

#ifdef USE_NLS
   /* Init NLS support */
   setlocale(LC_MESSAGES, "");
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
#endif   
   
   /* Init stat counters */
   init_stat(!curr);
   
   /* Process args... */
   while (opt < argc) {
      
      if (!strcmp(argv[opt], "-c")) {       /* Display cpu usage only		*/
	 cdf = D_CPU_ONLY;
	 opt++;
      }
      
      else if (!strcmp(argv[opt], "-d")) {  /* Display disk utilization only	*/
	 cdf = D_DISK_ONLY;
	 opt++;
      }
      
      else if (!strcmp(argv[opt], "-o")) {  /* Output to specified file		*/
	 if (argv[opt + 1])
	   filename = argv[++opt];
	 else
	   usage(argv[0]);
	 opt++;
      }
      
      else if (!i) {
	 interval = atol(argv[opt++]);
	 if (interval < 1)
	   usage(argv[0]);
	 count = -1;
	 i = 1;
      }
      
      else {
	 count = atol(argv[opt++]);
	 if (count < 1)
	   usage(argv[0]);
      }
   }

   /* Check if output must be sent to a file */
   if (filename) {
      
      /* Open output file */
      if ((fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC,
		     S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
	 perror("open");
	 exit(2);
      }
      
      /* stdout data is now sent to the file */
      if (dup2(fd, STDOUT_FILENO) < 0) {
	 perror("dup2");
	 exit(2);
      }
   }
   
   /* Set a handler for SIGALRM */
   alarm_handler(0);

   /* Main loop */
   do {
      /* Read uptime */
      read_uptime(curr);
      
      /* Read kernel statistics */
      read_stat(curr);

      /* Print results */
      print_stat(curr, cdf);
      fflush(stdout);
      
      if (count > 0)
	count--;

      if (count != 0)
	pause();
      curr ^= 1;
      
   }
   while (count != 0);

   return 0;
}

